from __future__ import annotations
from typing import Dict, List
import datetime
import msgpack
import webcface.func_info
import webcface.view_base
import webcface.field
import webcface.log_handler


class MessageBase:
    kind_def = -1
    kind: int
    msg: dict

    def __init__(self, kind: int, msg: dict):
        self.kind = kind
        self.msg = msg


def time_to_int(t: datetime.datetime) -> int:
    return int(t.timestamp() * 1000)


def int_to_time(t: int) -> datetime.datetime:
    return datetime.datetime.fromtimestamp(t / 1000)


class SyncInit(MessageBase):
    kind_def = 80

    def __init__(self, msg: dict) -> None:
        super().__init__(self.kind_def, msg)

    @staticmethod
    def new(M: str, l: str, v: str) -> SyncInit:
        return SyncInit.new_full(M, 0, l, v, "")

    @staticmethod
    def new_full(M: str, m: int, l: str, v: str, a: str) -> SyncInit:
        return SyncInit({"M": M, "m": m, "l": l, "v": v, "a": a})

    @property
    def member_name(self) -> str:
        return self.msg["M"]

    @property
    def member_id(self) -> int:
        return self.msg["m"]

    @property
    def lib_name(self) -> str:
        return self.msg["l"]

    @property
    def lib_ver(self) -> str:
        return self.msg["v"]

    @property
    def addr(self) -> str:
        return self.msg["a"]


class SvrVersion(MessageBase):
    kind_def = 88

    def __init__(self, msg: dict) -> None:
        super().__init__(self.kind_def, msg)

    @staticmethod
    def new(n: str, v: str) -> SvrVersion:
        return SvrVersion({"n": n, "v": v})

    @property
    def svr_name(self) -> str:
        return self.msg["n"]

    @property
    def ver(self) -> str:
        return self.msg["v"]


class Ping(MessageBase):
    kind_def = 89

    def __init__(self, msg: dict) -> None:
        super().__init__(self.kind_def, msg)

    @staticmethod
    def new() -> Ping:
        return Ping({})


class PingStatus(MessageBase):
    kind_def = 90

    def __init__(self, msg: dict) -> None:
        super().__init__(self.kind_def, msg)

    @staticmethod
    def new(s: Dict[int, int]) -> PingStatus:
        return PingStatus({"s": s})

    @property
    def status(self) -> Dict[int, int]:
        return self.msg["s"]


class PingStatusReq(MessageBase):
    kind_def = 91

    def __init__(self, msg: dict) -> None:
        super().__init__(self.kind_def, msg)

    @staticmethod
    def new() -> PingStatusReq:
        return PingStatusReq({})


class Sync(MessageBase):
    kind_def = 87

    def __init__(self, msg: dict) -> None:
        super().__init__(self.kind_def, msg)

    @staticmethod
    def new() -> Sync:
        return Sync({"m": 0, "t": time_to_int(datetime.datetime.now())})

    @staticmethod
    def new_full(m: int, t: int) -> Sync:
        return Sync({"m": m, "t": t})

    @property
    def member_id(self) -> int:
        return self.msg["m"]

    @property
    def time(self) -> datetime.datetime:
        return int_to_time(self.msg["t"])


class Value(MessageBase):
    kind_def = 0

    def __init__(self, msg: dict) -> None:
        super().__init__(self.kind_def, msg)

    @staticmethod
    def new(f: str, d: List[float]) -> Value:
        return Value({"f": f, "d": d})

    @property
    def field(self) -> str:
        return self.msg["f"]

    @property
    def data(self) -> List[float]:
        return self.msg["d"]


class ValueReq(MessageBase):
    kind_def = 40

    def __init__(self, msg: dict) -> None:
        super().__init__(self.kind_def, msg)

    @staticmethod
    def new(m: str, f: str, i: int) -> ValueReq:
        return ValueReq({"M": m, "f": f, "i": i})

    @property
    def member(self) -> str:
        return self.msg["M"]

    @property
    def field(self) -> str:
        return self.msg["f"]

    @property
    def req_id(self) -> int:
        return self.msg["i"]


class ValueRes(MessageBase):
    kind_def = 60

    def __init__(self, msg: dict) -> None:
        super().__init__(self.kind_def, msg)

    @staticmethod
    def new(i: int, f: str, d: List[float]) -> ValueRes:
        return ValueRes({"i": i, "f": f, "d": d})

    @property
    def req_id(self) -> int:
        return self.msg["i"]

    @property
    def sub_field(self) -> str:
        return self.msg["f"]

    @property
    def data(self) -> List[float]:
        return self.msg["d"]


class ValueEntry(MessageBase):
    kind_def = 20

    def __init__(self, msg: dict) -> None:
        super().__init__(self.kind_def, msg)

    @staticmethod
    def new(m: int, f: str) -> ValueEntry:
        return ValueEntry({"m": m, "f": f})

    @property
    def member_id(self) -> int:
        return self.msg["m"]

    @property
    def field(self) -> str:
        return self.msg["f"]


class Text(MessageBase):
    kind_def = 1

    def __init__(self, msg: dict) -> None:
        super().__init__(self.kind_def, msg)

    @staticmethod
    def new(f: str, d: str) -> Text:
        return Text({"f": f, "d": d})

    @property
    def field(self) -> str:
        return self.msg["f"]

    @property
    def data(self) -> str:
        return self.msg["d"]


class TextReq(MessageBase):
    kind_def = 41

    def __init__(self, msg: dict) -> None:
        super().__init__(self.kind_def, msg)

    @staticmethod
    def new(m: str, f: str, i: int) -> TextReq:
        return TextReq({"M": m, "f": f, "i": i})

    @property
    def member(self) -> str:
        return self.msg["M"]

    @property
    def field(self) -> str:
        return self.msg["f"]

    @property
    def req_id(self) -> int:
        return self.msg["i"]


class TextRes(MessageBase):
    kind_def = 61

    def __init__(self, msg: dict) -> None:
        super().__init__(self.kind_def, msg)

    @staticmethod
    def new(i: int, f: str, d: str) -> TextRes:
        return TextRes({"i": i, "f": f, "d": d})

    @property
    def req_id(self) -> int:
        return self.msg["i"]

    @property
    def sub_field(self) -> str:
        return self.msg["f"]

    @property
    def data(self) -> str:
        return self.msg["d"]


class TextEntry(MessageBase):
    kind_def = 21

    def __init__(self, msg: dict) -> None:
        super().__init__(self.kind_def, msg)

    @staticmethod
    def new(m: int, f: str) -> TextEntry:
        return TextEntry({"m": m, "f": f})

    @property
    def member_id(self) -> int:
        return self.msg["m"]

    @property
    def field(self) -> str:
        return self.msg["f"]


def vb_to_vd(vb: Dict[str, webcface.view_base.ViewComponentBase]) -> dict:
    """ViewComponentBaseクラスからメッセージに変換"""
    vd = {}
    for i, b in vb.items():
        vd[i] = {
            "t": b._type,
            "x": b._text,
            "L": None if b._on_click_func is None else b._on_click_func._member,
            "l": None if b._on_click_func is None else b._on_click_func._field,
            "c": b._text_color,
            "b": b._bg_color,
        }
    return vd


def vd_to_vb(vd: dict) -> Dict[str, webcface.view_base.ViewComponentBase]:
    """メッセージからViewComponentBaseクラスに変換"""
    vb = {}
    for i, d in vd.items():
        vb[i] = webcface.view_base.ViewComponentBase(
            type=d["t"],
            text=d["x"],
            on_click=(
                None
                if d["L"] is None or d["l"] is None
                else webcface.field.FieldBase(d["L"], d["l"])
            ),
            text_color=d["c"],
            bg_color=d["b"],
        )
    return vb


class View(MessageBase):
    kind_def = 3

    def __init__(self, msg: dict) -> None:
        super().__init__(self.kind_def, msg)

    @staticmethod
    def new(f: str, d: Dict[str, webcface.view_base.ViewComponentBase], l: int) -> View:
        return View({"f": f, "d": vb_to_vd(d), "l": l})

    @property
    def field(self) -> str:
        return self.msg["f"]

    @property
    def data(self) -> Dict[str, webcface.view_base.ViewComponentBase]:
        return vd_to_vb(self.msg["d"])

    @property
    def length(self) -> int:
        return self.msg["l"]


class ViewReq(MessageBase):
    kind_def = 43

    def __init__(self, msg: dict) -> None:
        super().__init__(self.kind_def, msg)

    @staticmethod
    def new(m: str, f: str, i: int) -> ViewReq:
        return ViewReq({"M": m, "f": f, "i": i})

    @property
    def member(self) -> str:
        return self.msg["M"]

    @property
    def field(self) -> str:
        return self.msg["f"]

    @property
    def req_id(self) -> int:
        return self.msg["i"]


class ViewRes(MessageBase):
    kind_def = 63

    def __init__(self, msg: dict) -> None:
        super().__init__(self.kind_def, msg)

    @staticmethod
    def new(
        i: int, f: str, d: Dict[str, webcface.view_base.ViewComponentBase], l: int
    ) -> ViewRes:
        return ViewRes({"i": i, "f": f, "d": vb_to_vd(d), "l": l})

    @property
    def req_id(self) -> int:
        return self.msg["i"]

    @property
    def sub_field(self) -> str:
        return self.msg["f"]

    @property
    def data_diff(self) -> Dict[str, webcface.view_base.ViewComponentBase]:
        return vd_to_vb(self.msg["d"])

    @property
    def length(self) -> int:
        return self.msg["l"]


class ViewEntry(MessageBase):
    kind_def = 23

    def __init__(self, msg: dict) -> None:
        super().__init__(self.kind_def, msg)

    @staticmethod
    def new(m: int, f: str) -> ViewEntry:
        return ViewEntry({"m": m, "f": f})

    @property
    def member_id(self) -> int:
        return self.msg["m"]

    @property
    def field(self) -> str:
        return self.msg["f"]


class FuncInfo(MessageBase):
    kind_def = 84

    def __init__(self, msg: dict) -> None:
        super().__init__(self.kind_def, msg)

    @staticmethod
    def new(f: str, fi: webcface.func_info.FuncInfo) -> FuncInfo:
        return FuncInfo.new_full(0, f, fi)

    @staticmethod
    def new_full(m: int, f: str, fi: webcface.func_info.FuncInfo) -> FuncInfo:
        ad = []
        for a in fi.args:
            ad.append(
                {
                    "n": a.name,
                    "t": a.type,
                    "i": a.init,
                    "m": a.min,
                    "x": a.max,
                    "o": a.option,
                }
            )
        return FuncInfo({"m": m, "f": f, "r": fi.return_type, "a": ad})

    @property
    def member_id(self) -> int:
        return self.msg["m"]

    @property
    def field(self) -> str:
        return self.msg["f"]

    @property
    def func_info(self) -> webcface.func_info.FuncInfo:
        args = []
        for a in self.msg["a"]:
            args.append(
                webcface.func_info.Arg(
                    name=a["n"],
                    type=a["t"],
                    init=a["i"],
                    min=a["m"],
                    max=a["x"],
                    option=a["o"],
                )
            )
        return webcface.func_info.FuncInfo(None, self.msg["r"], args, False)


class Call(MessageBase):
    kind_def = 81

    def __init__(self, msg: dict) -> None:
        super().__init__(self.kind_def, msg)

    @staticmethod
    def new(i: int, c: int, r: int, f: str, a: List[float | bool | str]) -> Call:
        return Call({"i": i, "c": c, "r": r, "f": f, "a": a})

    @property
    def caller_id(self) -> int:
        return self.msg["i"]

    @property
    def caller_member_id(self) -> int:
        return self.msg["c"]

    @property
    def target_member_id(self) -> int:
        return self.msg["r"]

    @property
    def field(self) -> str:
        return self.msg["f"]

    @property
    def args(self) -> List[float | bool | str]:
        return self.msg["a"]


class CallResponse(MessageBase):
    kind_def = 82

    def __init__(self, msg: dict) -> None:
        super().__init__(self.kind_def, msg)

    @staticmethod
    def new(i: int, c: int, s: bool) -> CallResponse:
        return CallResponse({"i": i, "c": c, "s": s})

    @property
    def caller_id(self) -> int:
        return self.msg["i"]

    @property
    def caller_member_id(self) -> int:
        return self.msg["c"]

    @property
    def started(self) -> bool:
        return self.msg["s"]


class CallResult(MessageBase):
    kind_def = 83

    def __init__(self, msg: dict) -> None:
        super().__init__(self.kind_def, msg)

    @staticmethod
    def new(i: int, c: int, e: bool, r: float | bool | str) -> CallResult:
        return CallResult({"i": i, "c": c, "e": e, "r": r})

    @property
    def caller_id(self) -> int:
        return self.msg["i"]

    @property
    def caller_member_id(self) -> int:
        return self.msg["c"]

    @property
    def is_error(self) -> bool:
        return self.msg["e"]

    @property
    def result(self) -> float | bool | str:
        return self.msg["r"]


class Log(MessageBase):
    kind_def = 85

    def __init__(self, msg: dict) -> None:
        super().__init__(self.kind_def, msg)

    @staticmethod
    def new(lls: List[webcface.log_handler.LogLine]) -> Log:
        return Log.new_full(0, lls)

    @staticmethod
    def new_full(m: int, lls: List[webcface.log_handler.LogLine]) -> Log:
        return Log(
            {
                "m": m,
                "l": [
                    {"v": ll.level, "t": time_to_int(ll.time), "m": ll.message}
                    for ll in lls
                ],
            }
        )

    @property
    def member_id(self) -> int:
        return self.msg["m"]

    @property
    def log(self) -> List[webcface.log_handler.LogLine]:
        return [
            webcface.log_handler.LogLine(l["v"], int_to_time(l["t"]), l["m"])
            for l in self.msg["l"]
        ]


class LogReq(MessageBase):
    kind_def = 86

    def __init__(self, msg: dict) -> None:
        super().__init__(self.kind_def, msg)

    @staticmethod
    def new(m: str) -> LogReq:
        return LogReq({"M": m})

    @property
    def member_name(self) -> str:
        return self.msg["M"]


# 受信する可能性のあるメッセージのリスト
message_classes_recv = [
    SyncInit,
    SvrVersion,
    Ping,
    PingStatus,
    Sync,
    ValueRes,
    ValueEntry,
    TextRes,
    TextEntry,
    ViewRes,
    ViewEntry,
    FuncInfo,
    Call,
    CallResponse,
    CallResult,
    Log,
]


def pack(msgs: List[MessageBase]) -> bytes:
    send_msgs: List[int | dict] = []
    for m in msgs:
        send_msgs.append(m.kind)
        send_msgs.append(m.msg)
    return msgpack.packb(send_msgs)


def unpack(packed: bytes) -> List[MessageBase]:
    unpack_obj = msgpack.unpackb(packed, strict_map_key=False)
    assert len(unpack_obj) % 2 == 0
    msg_ret = []
    for i in range(0, len(unpack_obj), 2):
        kind = unpack_obj[i]
        msg = unpack_obj[i + 1]
        assert isinstance(kind, int)
        assert isinstance(msg, dict)
        for C in message_classes_recv:
            if kind == C.kind_def:
                msg_ret.append(C(msg))
    return msg_ret
